#!/usr/bin/env python

#import smmp
import sys, copy,  random,  math,  numpy as np

from mpi4py import MPI 
#from algorithms import CanonicalMonteCarlo

class ParallelTempering(object):
    """Parallel tempering is a generalized Monte Carlo algorithm. A set of
    replica is simulated at different temperatures. Each replica performs a 
    canoncial Monte Carlo simulation. After a given number of sweeps, an exchange
    is attempted between neighboring replicas.
    """
    
    def __init__(self, T,  myUniverse):
        """Initialize ParallelTempering with a set of tempertures T.
        """
        self.kB = 1.98773e-3
        
        self._T = copy.deepcopy(T)
        self._numberOfReplica = len(T)

        # Calculate the number of processors available per replica
        numberOfProcessors = MPI.COMM_WORLD.Get_size()
        numberOfProcessorsPerReplica = numberOfProcessors // self._numberOfReplica
        numberOfProcessors = numberOfProcessorsPerReplica * self._numberOfReplica
        self._myReplica = MPI.COMM_WORLD.Get_rank() // numberOfProcessorsPerReplica

        # Create the replica internal communicators for the energy function
        groupWorld = MPI.COMM_WORLD.Get_group()
        group = [MPI.Group.Range_incl(groupWorld, [[i * numberOfProcessorsPerReplica, (i + 1) * numberOfProcessorsPerReplica - 1, 1], ]) for i in xrange(self._numberOfReplica)]
        comm = [MPI.COMM_WORLD.Create(g) for g in group]
        
        # Create the communicator needed to exchange information about replica
        ptGroup = MPI.Group.Incl(groupWorld, range(0, numberOfProcessors, numberOfProcessorsPerReplica))
        self._ptComm = MPI.COMM_WORLD.Create(ptGroup)
        
        self._localMaster = self._ptComm != MPI.COMM_NULL
        if self._localMaster:
            self._myPTRank = self._ptComm.Get_rank()
            
        self._oddCycle = False
        
        
    def exchange(self):
        """Collect temperatures and energies from all replicas, determine neighbor and do replica exchange move.
        
        The exchange step is the heart of parallel tempering. The probability of exchange between two 
        temperatures i and j is given by P_{ij} = \Delta \beta \Delta E, where \beta = 1/k_B T.
        """
        if self._localMaster:
#            energies = ptComm.allgather(self._myUniverse.energy())
            myE = [250,  255,  75,  125, 120, 185][self._myReplica]
            myT = [420,  370,  250,  275,  300,  330][self._myReplica]
            energies = self._ptComm.allgather(myE)
#            temperatures = ptComm.allgather(self._myUniverse.temperature())
            temperatures = self._ptComm.allgather(myT)
            
            temperatureRank = zip(temperatures,  np.arange(self._numberOfReplica))
            temperatureRank.sort()
            
            if (self._myPTRank % 2):
                if self._oddCycle:
                    rank0 = self._myPTRank
                else:
                    rank0 = self._myPTRank - 1
                
                rank1 = (rank0 + 1) % self._ptComm.Get_size()
                
                self._oddCycle = not self._oddCycle
                
                E0 = energies[temperatureRank[rank0][1]]
                E1 = energies[temperatureRank[rank1][1]]
                T0 = self._T[rank0]
                T1 = self._T[rank1]
                beta0 = 1.0 / (T0 * self.kB)
                beta1 = 1.0 / (T1 * self.kB)
                
                P = math.exp((beta0 - beta1) * (E0 - E1))
                print self._myPTRank,  P
                
                if P >= random.random(): # accept exchange
                    rank = self._ptComm.allgather(np.array([temperatureRank[rank1][1],  temperatureRank[rank0][1]]))
                else:
                    rank = self._ptComm.allgather(np.array([temperatureRank[rank0][1],  temperatureRank[rank1][1]]))
            else:
                random.random() # Do one random number to stay in sync with the other processes for the same replica.
                rank = self._ptComm.allgather(np.array([-1,  -1]))

            rank = np.array(rank[1::2]).flatten()
            print(rank)
            rankTemperature = zip(rank,  self._T)
            rankTemperature.sort()
            print self._myPTRank,  rankTemperature[self._myReplica][1]
#            self._myUniverse.setTemperature(rankTemperature[self._myReplica][1])
        else:
            random.random() # Do one random number to stay in sync with the master process for the same replica.
            

            
 
if __name__ == "__main__":
    
    n = int(sys.argv[1])
    T = [250,  275,  300,  330,  370,  420]
    
    myPT = ParallelTempering(T,  [])
    myPT.exchange()
    myPT.exchange()


        
        
        

         
